@@ -8,6 +8,9 @@
 #include "globals.h"
 #include "parser.h"
 #include "dfa.h"
+#include "print.h"
+#include "utf8_regexp.h"
+#include "utf16_regexp.h"
 
 namespace re2c
 {
@@ -83,6 +86,13 @@ uint RegExp::fixedLength()
 	return ~0u;
 }
 
+uint compile_goto(Ins *ins, Ins *i)
+{
+	i->i.tag = GOTO;
+	i->i.link = ins;
+	return 1;
+}
+
 const char *NullOp::type = "NullOp";
 
 void NullOp::calcSize(Char*)
@@ -95,7 +105,12 @@ uint NullOp::fixedLength()
 	return 0;
 }
 
-void NullOp::compile(Char*, Ins*)
+uint NullOp::compile(Char*, Ins*)
+{
+	return 0;
+}
+
+void NullOp::decompile()
 {
 	;
 }
@@ -105,131 +120,6 @@ void NullOp::split(CharSet&)
 	;
 }
 
-std::ostream& operator<<(std::ostream &o, const Range &r)
-{
-	if ((r.ub - r.lb) == 1)
-	{
-		prtCh(o, r.lb);
-	}
-	else
-	{
-		prtCh(o, r.lb);
-		o << "-";
-		prtCh(o, r.ub - 1);
-	}
-
-	return o << r.next;
-}
-
-Range *doUnion(Range *r1, Range *r2)
-{
-	Range *r, **rP = &r;
-
-	for (;;)
-	{
-		Range *s;
-
-		if (r1->lb <= r2->lb)
-		{
-			s = new Range(*r1);
-		}
-		else
-		{
-			s = new Range(*r2);
-		}
-
-		*rP = s;
-		rP = &s->next;
-
-		for (;;)
-		{
-			if (r1->lb <= r2->lb)
-			{
-				if (r1->lb > s->ub)
-					break;
-
-				if (r1->ub > s->ub)
-					s->ub = r1->ub;
-
-				if (!(r1 = r1->next))
-				{
-					uint ub = 0;
-
-					for (; r2 && r2->lb <= s->ub; r2 = r2->next)
-						ub = r2->ub;
-
-					if (ub > s->ub)
-						s->ub = ub;
-
-					*rP = r2;
-
-					return r;
-				}
-			}
-			else
-			{
-				if (r2->lb > s->ub)
-					break;
-
-				if (r2->ub > s->ub)
-					s->ub = r2->ub;
-
-				if (!(r2 = r2->next))
-				{
-					uint ub = 0;
-
-					for (; r1 && r1->lb <= s->ub; r1 = r1->next)
-						ub = r1->ub;
-
-					if (ub > s->ub)
-						s->ub = ub;
-
-					*rP = r1;
-
-					return r;
-				}
-			}
-		}
-	}
-
-	*rP = NULL;
-	return r;
-}
-
-Range *doDiff(Range *r1, Range *r2)
-{
-	Range *r, *s, **rP = &r;
-
-	for (; r1; r1 = r1->next)
-	{
-		uint lb = r1->lb;
-
-		for (; r2 && r2->ub <= r1->lb; r2 = r2->next)
-
-			;
-		for (; r2 && r2->lb < r1->ub; r2 = r2->next)
-		{
-			if (lb < r2->lb)
-			{
-				*rP = s = new Range(lb, r2->lb);
-				rP = &s->next;
-			}
-
-			if ((lb = r2->ub) >= r1->ub)
-				goto noMore;
-		}
-
-		*rP = s = new Range(lb, r1->ub);
-		rP = &s->next;
-
-noMore:
-		;
-	}
-
-	*rP = NULL;
-	return r;
-}
-
 MatchOp *merge(MatchOp *m1, MatchOp *m2)
 {
 	if (!m1)
@@ -263,27 +153,44 @@ uint MatchOp::fixedLength()
 	return 1;
 }
 
-void MatchOp::compile(Char *rep, Ins *i)
+uint MatchOp::compile(Char *rep, Ins *i)
 {
-	i->i.tag = CHAR;
-	i->i.link = &i[size];
-	Ins *j = &i[1];
-	uint bump = size;
-
-	for (Range *r = match; r; r = r->next)
+	if (ins_cache)
+		return compile_goto(ins_cache, i);
+	else
 	{
-		for (uint c = r->lb; c < r->ub; ++c)
+		ins_cache = i;
+
+		i->i.tag = CHAR;
+		i->i.link = &i[size];
+		Ins *j = &i[1];
+		uint bump = size;
+
+		for (Range *r = match; r; r = r->next)
 		{
-			if (rep[c] == c)
+			for (uint c = r->lb; c < r->ub; ++c)
 			{
-				j->c.value = c;
-				j->c.bump = --bump;
-				j++;
+				if (rep[c] == c)
+				{
+					j->c.value = c;
+					j->c.bump = --bump;
+					j++;
+				}
 			}
 		}
+
+		if (ins_access == PRIVATE)
+			decompile();
+
+		return size;
 	}
 }
 
+void MatchOp::decompile()
+{
+	ins_cache = NULL;
+}
+
 void MatchOp::split(CharSet &s)
 {
 	for (Range *r = match; r; r = r->next)
@@ -351,6 +258,17 @@ RegExp *doAlt(RegExp *e1, RegExp *e2)
 	return new AltOp(e1, e2);
 }
 
+RegExp *doCat(RegExp *e1, RegExp *e2)
+{
+	if (!e1)
+		return e2;
+
+	if (!e2)
+		return e1;
+
+	return new CatOp(e1, e2);
+}
+
 RegExp *mkAlt(RegExp *e1, RegExp *e2)
 {
 	AltOp *a;
@@ -399,15 +317,37 @@ uint AltOp::fixedLength()
 	return l1;
 }
 
-void AltOp::compile(Char *rep, Ins *i)
+uint AltOp::compile(Char *rep, Ins *i)
 {
-	i->i.tag = FORK;
-	Ins *j = &i[exp1->size + 1];
-	i->i.link = &j[1];
-	exp1->compile(rep, &i[1]);
-	j->i.tag = GOTO;
-	j->i.link = &j[exp2->size + 1];
-	exp2->compile(rep, &j[1]);
+	if (ins_cache)
+		return compile_goto(ins_cache, i);
+	else
+	{
+		ins_cache = i;
+
+		i->i.tag = FORK;
+		const uint sz1 = exp1->compile(rep, &i[1]);
+		Ins * const j = &i[sz1 + 1];
+		i->i.link = &j[1];
+		j->i.tag = GOTO;
+		const uint sz2 = exp2->compile(rep, &j[1]);
+		j->i.link = &j[sz2 + 1];
+
+		if (ins_access == PRIVATE)
+			decompile();
+
+		return sz1 + sz2 + 2;
+	}
+}
+
+void AltOp::decompile()
+{
+	if (ins_cache)
+	{
+		exp1->decompile();
+		exp2->decompile();
+		ins_cache = NULL;
+	}
 }
 
 void AltOp::split(CharSet &s)
@@ -436,10 +376,32 @@ uint CatOp::fixedLength()
 	return ~0u;
 }
 
-void CatOp::compile(Char *rep, Ins *i)
+uint CatOp::compile(Char *rep, Ins *i)
 {
-	exp1->compile(rep, &i[0]);
-	exp2->compile(rep, &i[exp1->size]);
+	if (ins_cache)
+		return compile_goto(ins_cache, i);
+	else
+	{
+		ins_cache = i;
+
+		const uint sz1 = exp1->compile(rep, &i[0]);
+		const uint sz2 = exp2->compile(rep, &i[sz1]);
+
+		if (ins_access == PRIVATE)
+			decompile();
+
+		return sz1 + sz2;
+	}
+}
+
+void CatOp::decompile()
+{
+	if (ins_cache)
+	{
+		exp1->decompile();
+		exp2->decompile();
+		ins_cache = NULL;
+	}
 }
 
 void CatOp::split(CharSet &s)
@@ -456,12 +418,34 @@ void CloseOp::calcSize(Char *rep)
 	size = exp->size + 1;
 }
 
-void CloseOp::compile(Char *rep, Ins *i)
+uint CloseOp::compile(Char *rep, Ins *i)
 {
-	exp->compile(rep, &i[0]);
-	i += exp->size;
-	i->i.tag = FORK;
-	i->i.link = i - exp->size;
+	if (ins_cache)
+		return compile_goto(ins_cache, i);
+	else
+	{
+		ins_cache = i;
+
+		i += exp->compile(rep, &i[0]);
+		i->i.tag = FORK;
+		i->i.link = ins_cache;
+		++i;
+
+		const uint sz = i - ins_cache;
+		if (ins_access == PRIVATE)
+			decompile();
+
+		return sz;
+	}
+}
+
+void CloseOp::decompile()
+{
+	if (ins_cache)
+	{
+		exp->decompile();
+		ins_cache = NULL;
+	}
 }
 
 void CloseOp::split(CharSet &s)
@@ -485,32 +469,48 @@ void CloseVOp::calcSize(Char *rep)
 	}
 }
 
-void CloseVOp::compile(Char *rep, Ins *i)
+uint CloseVOp::compile(Char *rep, Ins *i)
 {
-	Ins *jumppoint;
-	int st;
-	jumppoint = i + ((1 + exp->size) * (max - min));
-
-	for (st = min; st < max; st++)
+	if (ins_cache)
+		return compile_goto(ins_cache, i);
+	else
 	{
-		i->i.tag = FORK;
-		i->i.link = jumppoint;
-		i++;
-		exp->compile(rep, &i[0]);
-		i += exp->size;
-	}
+		ins_cache = i;
 
-	for (st = 0; st < min; st++)
-	{
-		exp->compile(rep, &i[0]);
-		i += exp->size;
-
-		if (max < 0 && st == 0)
+		for (int st = min; st < max; st++)
 		{
+			const uint sz = exp->compile(rep, &i[1]);
 			i->i.tag = FORK;
-			i->i.link = i - exp->size;
-			i++;
+			i->i.link = ins_cache + (1 + sz) * (max - min);
+			i += sz + 1;
 		}
+
+		for (int st = 0; st < min; st++)
+		{
+			const uint sz = exp->compile(rep, &i[0]);
+			i += sz;
+			if (max < 0 && st == 0)
+			{
+				i->i.tag = FORK;
+				i->i.link = i - sz;
+				i++;
+			}
+		}
+
+		const uint sz = i - ins_cache;
+		if (ins_access == PRIVATE)
+			decompile();
+
+		return sz;
+	}
+}
+
+void CloseVOp::decompile()
+{
+	if (ins_cache)
+	{
+		exp->decompile();
+		ins_cache = NULL;
 	}
 }
 
@@ -583,17 +583,17 @@ uint Scanner::unescape(SubStr &s) const
 			}
 
 			uint l = 0;
-						
 			if (s.str[0] == '0')
 			{
 				l++;
 				if (s.str[1] == '0')
 				{
 					l++;
-					if (s.str[2] == '0' || (s.str[2] == '1' && uFlag))
+					if (s.str[2] == '0' || (s.str[2] == '1' && encoding.szCodePoint() == 4))
 					{
 						l++;
-						if (uFlag) {
+						if (encoding.szCodePoint() == 4)
+						{
 							const char *u3 = strchr(hex, tolower(s.str[2]));
 							const char *u4 = strchr(hex, tolower(s.str[3]));
 							if (u3 && u4)
@@ -657,7 +657,7 @@ uint Scanner::unescape(SubStr &s) const
 				       + (uint)((p4 - hex))
 				       + ucb;
 	
-				if (v >= nRealChars)
+				if (v >= encoding.nCodePoints())
 				{
 					fatal(s.ofs(),
 						c == 'X'
@@ -735,7 +735,7 @@ std::string& Scanner::unescape(SubStr& str_in, std::string& str_out) const
 
 Range * Scanner::getRange(SubStr &s) const
 {
-	uint lb = unescape(s), ub, xlb, xub, c;
+	uint lb = unescape(s), ub;
 
 	if (s.len < 2 || *s.str != '-')
 	{
@@ -746,39 +746,31 @@ Range * Scanner::getRange(SubStr &s) const
 		s.len--;
 		s.str++;
 		ub = unescape(s);
-
 		if (ub < lb)
 		{
 			uint tmp = lb;
 			lb = ub;
 			ub = tmp;
 		}
-		
-		xlb = xlat(lb);
-		xub = xlat(ub);
-		
-		for(c = lb; c <= ub; c++)
-		{
-			if (!(xlb <= xlat(c) && xlat(c) <= ub))
-			{
-				/* range doesn't work */
-				Range * r = new Range(xlb, xlb + 1);
-				for (c = lb + 1; c <= ub; c++)
-				{
-					r = doUnion(r, new Range(xlat(c), xlat(c) + 1));
-				}
-				return r;
-			}
-		}
 	}
 
-	return new Range(xlat(lb), xlat(ub) + 1);
+	Range * r = encoding.encodeRange(lb, ub);
+	if (r == NULL)
+		fatalf("Bad code point range: '0x%X - 0x%X'", lb, ub);
+	return r;
 }
 
-RegExp * Scanner::matchChar(uint c) const
+RegExp * Scanner::matchSymbol(uint c) const
 {
-	uint xc = xlat(c);
-	return new MatchOp(new Range(xc, xc + 1));
+	if (!encoding.encode(c))
+		fatalf("Bad code point: '0x%X'", c);
+
+	if (encoding.is(Enc::UTF16))
+		return UTF16Symbol(c);
+	else if (encoding.is(Enc::UTF8))
+		return UTF8Symbol(c);
+	else
+		return new MatchOp(new Range(c, c + 1));
 }
 
 RegExp * Scanner::strToRE(SubStr s) const
@@ -789,10 +781,10 @@ RegExp * Scanner::strToRE(SubStr s) const
 	if (s.len == 0)
 		return new NullOp;
 
-	RegExp *re = matchChar(unescape(s));
+	RegExp *re = matchSymbol(unescape(s));
 
 	while (s.len > 0)
-		re = new CatOp(re, matchChar(unescape(s)));
+		re = new CatOp(re, matchSymbol(unescape(s)));
 
 	return re;
 }
@@ -811,13 +803,13 @@ RegExp * Scanner::strToCaseInsensitiveRE(SubStr s) const
 
 	if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
 	{
-		reL = matchChar(tolower(c));
-		reU = matchChar(toupper(c));
+		reL = matchSymbol(tolower(c));
+		reU = matchSymbol(toupper(c));
 		re = mkAlt(reL, reU);
 	}
 	else
 	{
-		re = matchChar(c);
+		re = matchSymbol(c);
 	}
 
 	while (s.len > 0)
@@ -826,19 +818,38 @@ RegExp * Scanner::strToCaseInsensitiveRE(SubStr s) const
 
 		if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
 		{
-			reL = matchChar(tolower(c));
-			reU = matchChar(toupper(c));
+			reL = matchSymbol(tolower(c));
+			reU = matchSymbol(toupper(c));
 			re = new CatOp(re, mkAlt(reL, reU));
 		}
 		else
 		{
-			re = new CatOp(re, matchChar(c));
+			re = new CatOp(re, matchSymbol(c));
 		}
 	}
 
 	return re;
 }
 
+Range * Scanner::mkRange(SubStr &s) const
+{
+	Range *r = getRange(s);
+	while (s.len > 0)
+		r = doUnion(r, getRange(s));
+
+	return r;
+}
+
+RegExp * Scanner::matchSymbolRange(Range * r) const
+{
+	if (encoding.is(Enc::UTF16))
+		return UTF16Range(r);
+	else if (encoding.is(Enc::UTF8))
+		return UTF8Range(r);
+	else
+		return new MatchOp(r);
+}
+
 RegExp * Scanner::ranToRE(SubStr s) const
 {
 	s.len -= 2;
@@ -847,62 +858,54 @@ RegExp * Scanner::ranToRE(SubStr s) const
 	if (s.len == 0)
 		return new NullOp;
 
-	Range *r = getRange(s);
-
-	while (s.len > 0)
-		r = doUnion(r, getRange(s));
-
-	return new MatchOp(r);
-}
-
-RegExp * Scanner::getAnyRE() const
-{
-	if (eFlag)
-	{
-		return ranToRE(SubStr("[\\000-\\377]"));
-	}
-	else
-	{
-		return new MatchOp(new Range(0, nRealChars));
-	}
+	return matchSymbolRange(mkRange(s));
 }
 
 RegExp * Scanner::invToRE(SubStr s) const
 {
-	s.len--;
-	s.str++;
-	
-	RegExp * any = getAnyRE();
+	s.len -= 3;
+	s.str += 2;
 
-	if (s.len <= 2)
-	{
-		return any;
-	}
+	Range * full = encoding.fullRange();
 
-	RegExp * ran = ranToRE(s);
-	RegExp * inv = mkDiff(any, ran);
-	
-	delete ran;
-	delete any;
-	
-	return inv;
+	Range * r = s.len == 0
+		? full
+		: doDiff(full, mkRange (s));
+
+	return matchSymbolRange(r);
 }
 
 RegExp * Scanner::mkDot() const
 {
-	RegExp * any = getAnyRE();
-	RegExp * ran = matchChar('\n');
-	RegExp * inv = mkDiff(any, ran);
-	
-	delete ran;
-	delete any;
-	
-	return inv;
+	Range * full = encoding.fullRange();
+	uint c = '\n';
+	if (!encoding.encode(c))
+		fatalf("Bad code point: '0x%X'", c);
+	Range * ran = new Range(c, c + 1);
+	Range * inv = doDiff(full, ran);
+
+	return matchSymbolRange(inv);
+}
+
+/*
+ * Create a byte range that includes all possible input characters.
+ * This may include characters, which do not map to any valid symbol
+ * in current encoding. For encodings, which directly map symbols to
+ * input characters (ASCII, EBCDIC, UTF-32), it equals [^]. For other
+ * encodings (UTF-16, UTF-8), [^] and this range are different.
+ *
+ * Also note that default range doesn't respect encoding policy
+ * (the way invalid code points are treated).
+ */
+RegExp * Scanner::mkDefault() const
+{
+	Range * def = new Range(0, encoding.nCodeUnits());
+	return new MatchOp(def);
 }
 
 const char *RuleOp::type = "RuleOp";
 
-RuleOp::RuleOp(RegExp *e, RegExp *c, Token *t, uint a)
+RuleOp::RuleOp(RegExp *e, RegExp *c, Token *t, uint a, InsAccess access)
 	: exp(e)
 	, ctx(c)
 	, ins(NULL)
@@ -910,13 +913,13 @@ RuleOp::RuleOp(RegExp *e, RegExp *c, Token *t, uint a)
 	, code(t)
 	, line(0)
 {
-	;
+	ins_access = access;
 }
 
 RuleOp* RuleOp::copy(uint a) const
 {
 	Token *token = new Token(*code);
-	return new RuleOp(exp, ctx, token, a);
+	return new RuleOp(exp, ctx, token, a, ins_access);
 }
 
 void RuleOp::calcSize(Char *rep)
@@ -926,21 +929,42 @@ void RuleOp::calcSize(Char *rep)
 	size = exp->size + (ctx->size ? ctx->size + 2 : 1);
 }
 
-void RuleOp::compile(Char *rep, Ins *i)
+uint RuleOp::compile(Char *rep, Ins *i)
 {
-	ins = i;
-	exp->compile(rep, &i[0]);
-	i += exp->size;
-	if (ctx->size)
+	if (ins_cache)
+		return compile_goto(ins_cache, i);
+	else
 	{
-		i->i.tag = CTXT;
-		i->i.link = &i[1];
-		i++;
-		ctx->compile(rep, &i[0]);
-		i += ctx->size;
+		ins_cache = i;
+
+		i += exp->compile(rep, &i[0]);
+		if (ctx->size)
+		{
+			i->i.tag = CTXT;
+			i->i.link = &i[1];
+			++i;
+			i += ctx->compile(rep, &i[0]);
+		}
+		i->i.tag = TERM;
+		i->i.link = this;
+		++i;
+
+		const uint sz = i - ins_cache;
+		if (ins_access == PRIVATE)
+			decompile();
+
+		return sz;
+	}
+}
+
+void RuleOp::decompile()
+{
+	if (ins_cache)
+	{
+		exp->decompile();
+		ctx->decompile();
+		ins_cache = NULL;
 	}
-	i->i.tag = TERM;
-	i->i.link = this;
 }
 
 void RuleOp::split(CharSet &s)
@@ -996,19 +1020,19 @@ CharSet::CharSet()
 	: fix(0)
 	, freeHead(0)
 	, freeTail(0)
-	, rep(new CharPtr[nRealChars])
-	, ptn(new CharPtn[nRealChars])
+	, rep(new CharPtr[encoding.nCodeUnits()])
+	, ptn(new CharPtn[encoding.nCodeUnits()])
 {
-	for (uint j = 0; j < nRealChars; ++j)
+	for (uint j = 0; j < encoding.nCodeUnits(); ++j)
 	{
 		rep[j] = &ptn[0];
-		ptn[j].nxt = &ptn[j + 1]; /* wrong for j=nRealChars but will be corrected below */
+		ptn[j].nxt = &ptn[j + 1]; /* wrong for j=encoding.nCodeUnits() - 1 but will be corrected below */
 		ptn[j].card = 0;
 	}
 
 	freeHead = &ptn[1];
-	*(freeTail = &ptn[nRealChars - 1].nxt) = NULL;
-	ptn[0].card = nRealChars;
+	*(freeTail = &ptn[encoding.nCodeUnits() - 1].nxt) = NULL;
+	ptn[0].card = encoding.nCodeUnits();
 	ptn[0].nxt = NULL;
 }
 	
@@ -1018,46 +1042,46 @@ CharSet::~CharSet()
 	delete[] ptn;
 }
 
-DFA* genCode(RegExp *re)
+smart_ptr<DFA> genCode(RegExp *re)
 {
 	CharSet cs;
-	uint j;
-
 	re->split(cs);
-	/*
-	    for(uint k = 0; k < nChars;){
-		for(j = k; ++k < nRealChars && cs.rep[k] == cs.rep[j];);
-		printSpan(cerr, j, k);
-		cerr << "\t" << cs.rep[j] - &cs.ptn[0] << endl;
-	    }
-	*/
-	Char *rep = new Char[nRealChars];
+	
+	/*for(uint k = 0; k < encoding.nCodeUnits();)
+	{
+		uint j;
+		for(j = k; ++k < encoding.nCodeUnits() && cs.rep[k] == cs.rep[j];);
+		printSpan(std::cerr, j, k);
+		std::cerr << "\t" << cs.rep[j] - &cs.ptn[0] << std::endl;
+	}*/
+	
+	Char *rep = new Char[encoding.nCodeUnits()];
 
-	for (j = 0; j < nRealChars; ++j)
+	for (uint j = 0; j < encoding.nCodeUnits(); ++j)
 	{
 		if (!cs.rep[j]->nxt)
 			cs.rep[j]->nxt = &cs.ptn[j];
 
-		rep[j] = (Char) (cs.rep[j]->nxt - &cs.ptn[0]);
+		rep[j] = cs.rep[j]->nxt - &cs.ptn[0];
 	}
 
 	re->calcSize(rep);
 	Ins *ins = new Ins[re->size + 1];
 	memset(ins, 0, (re->size + 1)*sizeof(Ins));
-	re->compile(rep, ins);
-	Ins *eoi = &ins[re->size];
+	const uint size = re->compile(rep, ins);
+	Ins *eoi = &ins[size];
 	eoi->i.tag = GOTO;
 	eoi->i.link = eoi;
 
 	optimize(ins);
 
 	/*
-	for (const Ins *inst = &ins[0]; inst < &ins[re->size]; ) {
+	for (const Ins *inst = &ins[0]; inst < &ins[size]; ) {
 		inst = showIns(std::cout, *inst, ins[0]);
 	}
 	*/
 
-	for (j = 0; j < re->size;)
+	for (uint j = 0; j < size;)
 	{
 		unmark(&ins[j]);
 
@@ -1071,7 +1095,7 @@ DFA* genCode(RegExp *re)
 		}
 	}
 
-	return new DFA(ins, re->size, 0, nRealChars, rep);
+	return make_smart_ptr(new DFA(ins, size, 0, encoding.nCodeUnits(), rep));
 }
 
 } // end namespace re2c
