@@ -10,56 +10,14 @@
 #include <string>
 #include "token.h"
 #include "ins.h"
+#include "free_list.h"
 #include "globals.h"
+#include "range.h"
+#include "smart_ptr.h"
 
 namespace re2c
 {
 
-template<class _Ty>
-class free_list: protected std::set<_Ty>
-{
-public:
-	typedef typename std::set<_Ty>::iterator   iterator;
-	typedef typename std::set<_Ty>::size_type  size_type;
-	typedef typename std::set<_Ty>::key_type   key_type;
-	
-	free_list(): in_clear(false)
-	{
-	}
-	
-	using std::set<_Ty>::insert;
-
-	size_type erase(const key_type& key)
-	{
-		if (!in_clear)
-		{
-			return std::set<_Ty>::erase(key);
-		}
-		return 0;
-	}
-	
-	void clear()
-	{
-		in_clear = true;
-
-		for(iterator it = this->begin(); it != this->end(); ++it)
-		{
-			delete *it;
-		}
-		std::set<_Ty>::clear();
-		
-		in_clear = false;
-	}
-
-	~free_list()
-	{
-		clear();
-	}
-
-protected:
-	bool in_clear;
-};
-
 typedef struct extop
 {
 	char op;
@@ -89,50 +47,43 @@ struct CharSet
 	CharPtn	*ptn;
 };
 
-class Range
-{
-
-public:
-	Range	*next;
-	uint	lb, ub;		// [lb,ub)
-
-	static free_list<Range*> vFreeList;
-
-public:
-	Range(uint l, uint u) : next(NULL), lb(l), ub(u)
-	{
-		vFreeList.insert(this);
-	}
-
-	Range(Range &r) : next(NULL), lb(r.lb), ub(r.ub)
-	{
-		vFreeList.insert(this);
-	}
-
-	~Range()
-	{
-		vFreeList.erase(this);
-	}
-
-	friend std::ostream& operator<<(std::ostream&, const Range&);
-	friend std::ostream& operator<<(std::ostream&, const Range*);
-};
-
-inline std::ostream& operator<<(std::ostream &o, const Range *r)
-{
-	return r ? o << *r : o;
-}
-
 class RegExp
 {
 
 public:
 	uint	size;
-	
+
+	/*
+	 * There're several different cases when the same regexp
+	 * can be used multiple times:
+	 * 	1) named definitions, e.g. digit = [0-9];
+	 * 	2) counted repetition, e.g. "a"{3}, "a"{3,}, "a"{3,5}
+	 * 	3) multiple DFA's sharing the same regexp , e.g. <c1, c2> "abc" { }
+	 * 	3) common suffixes, e.g. suffix [\x80-\xBF] in UTF-8 ranges
+	 * In cases 1-3, regexp must be recompiled each time it's reused.
+	 * In case 4, regexp should be compiled only once, and instructions
+	 * should be shared in order to reduce space.
+	 *
+	 * Note: if regexp must always be recompiled, it doesn't imply that
+	 * parts of this regexp must always be recompiled. It only forces
+	 * the compilation function to traverse the regexp after compilation
+	 * and reset compilation cache for each sub-regexp. E.g., for a regexp
+	 * [^]{3} in UTF-8 mode, each of sub-regexps [^] will have common suffix
+	 * [\x80-\xBF] factored out, but they won't share instructions.
+	 */
+	Ins*	ins_cache; /* if non-NULL, points to compiled instructions */
+	enum InsAccess
+		{ SHARED
+		, PRIVATE
+		} ins_access;
+
 	static free_list<RegExp*> vFreeList;
 
 public:
-	RegExp() : size(0)
+	RegExp()
+		: size(0)
+		, ins_cache(NULL)
+		, ins_access(SHARED)
 	{
 		vFreeList.insert(this);
 	}
@@ -151,7 +102,8 @@ public:
 	virtual void split(CharSet&) = 0;
 	virtual void calcSize(Char*) = 0;
 	virtual uint fixedLength();
-	virtual void compile(Char*, Ins*) = 0;
+	virtual uint compile(Char*, Ins*) = 0;
+	virtual void decompile() = 0;
 	virtual void display(std::ostream&) const = 0;
 	friend std::ostream& operator<<(std::ostream&, const RegExp&);
 	friend std::ostream& operator<<(std::ostream&, const RegExp*);
@@ -183,7 +135,8 @@ public:
 	void split(CharSet&);
 	void calcSize(Char*);
 	uint fixedLength();
-	void compile(Char*, Ins*);
+	uint compile(Char*, Ins*);
+	void decompile();
 	void display(std::ostream &o) const
 	{
 		o << "_";
@@ -210,7 +163,8 @@ public:
 	void split(CharSet&);
 	void calcSize(Char*);
 	uint fixedLength();
-	void compile(Char*, Ins*);
+	uint compile(Char*, Ins*);
+	void decompile();
 	void display(std::ostream&) const;
 
 #ifdef PEDANTIC
@@ -245,7 +199,7 @@ public:
 	uint     line;
 
 public:
-	RuleOp(RegExp*, RegExp*, Token*, uint);
+	RuleOp(RegExp*, RegExp*, Token*, uint, InsAccess);
 
 	~RuleOp()
 	{
@@ -259,7 +213,8 @@ public:
 
 	void split(CharSet&);
 	void calcSize(Char*);
-	void compile(Char*, Ins*);
+	uint compile(Char*, Ins*);
+	void decompile();
 	void display(std::ostream &o) const
 	{
 		o << exp << "/" << ctx << ";";
@@ -304,6 +259,7 @@ public:
 	const RuleOp& op;
 };
 
+RegExp *doAlt(RegExp*, RegExp*);
 RegExp *mkAlt(RegExp*, RegExp*);
 
 class AltOp: public RegExp
@@ -330,7 +286,8 @@ public:
 	void split(CharSet&);
 	void calcSize(Char*);
 	uint fixedLength();
-	void compile(Char*, Ins*);
+	uint compile(Char*, Ins*);
+	void decompile();
 	void display(std::ostream &o) const
 	{
 		o << exp1 << "|" << exp2;
@@ -354,6 +311,9 @@ private:
 #endif
 };
 
+RegExp *doCat(RegExp*, RegExp*);
+RegExp *mkCat(RegExp*, RegExp*);
+
 class CatOp: public RegExp
 {
 
@@ -378,7 +338,8 @@ public:
 	void split(CharSet&);
 	void calcSize(Char*);
 	uint fixedLength();
-	void compile(Char*, Ins*);
+	uint compile(Char*, Ins*);
+	void decompile();
 	void display(std::ostream &o) const
 	{
 		o << exp1 << exp2;
@@ -422,7 +383,8 @@ public:
 
 	void split(CharSet&);
 	void calcSize(Char*);
-	void compile(Char*, Ins*);
+	uint compile(Char*, Ins*);
+	void decompile();
 	void display(std::ostream &o) const
 	{
 		o << exp << "+";
@@ -460,6 +422,7 @@ public:
 		, min(lb)
 		, max(ub)
 	{
+		exp->ins_access = PRIVATE;
 	}
 
 	const char *typeOf()
@@ -469,7 +432,8 @@ public:
 
 	void split(CharSet&);
 	void calcSize(Char*);
-	void compile(Char*, Ins*);
+	uint compile(Char*, Ins*);
+	void decompile();
 	void display(std::ostream &o) const
 	{
 		o << exp << "+";
@@ -492,16 +456,17 @@ private:
 };
 
 typedef std::set<std::string>           CondList;
-typedef std::pair<unsigned, RegExp*>    NRegExp;
+typedef std::pair<int, RegExp*>         NRegExp;
 typedef std::map<std::string, NRegExp>  RegExpMap;
 typedef std::vector<std::string>        RegExpIndices;
 typedef std::list<RuleOp*>              RuleOpList;
 typedef std::pair<uint, std::string>    LineCode;
 typedef std::map<std::string, LineCode> SetupMap;
+typedef std::map<std::string, Token*>   DefaultMap;
 
 class DFA;
 
-extern DFA* genCode(RegExp*);
+extern smart_ptr<DFA> genCode(RegExp*);
 extern void genGetStateGoto(std::ostream&, uint&, uint);
 extern void genCondTable(std::ostream&, uint, const RegExpMap&);
 extern void genCondGoto(std::ostream&, uint, const RegExpMap&);
